"""Module for the base hyperparameter tuning class."""

from __future__ import annotations

import getpass
import inspect
import logging
import platform
import random
from datetime import datetime
from pathlib import Path
from typing import Callable, Hashable, Literal

import optuna
import vaex
from optuna.samplers import (
    BaseSampler,
    BruteForceSampler,
    CmaEsSampler,
    GridSampler,
    NSGAIIISampler,
    NSGAIISampler,
    PartialFixedSampler,
    QMCSampler,
    RandomSampler,
    TPESampler,
)
from optuna.samplers.nsgaii._child_generation_strategy import NSGAIIChildGenerationStrategy
from optuna.storages import RDBStorage
from optuna_dashboard import save_note
from tqdm.auto import tqdm

from mleko import __version__ as mleko_version
from mleko.cache.fingerprinters import (
    CallableSourceFingerprinter,
    OptunaPrunerFingerprinter,
    OptunaSamplerFingerprinter,
)
from mleko.cache.fingerprinters.dict_fingerprinter import DictFingerprinter
from mleko.dataset.data_schema import DataSchema
from mleko.model.base_model import HyperparametersType
from mleko.utils.custom_logger import CustomLogger

from .base_tuner import BaseTuner


logger = CustomLogger()
"""The logger for the module."""

OptimizeDirection = Literal["maximize", "minimize"]
"""Literal type for the direction of optimization."""


class OptunaTuner(BaseTuner):
    """Hyperparameter tuner using `Optuna`."""

    def __init__(
        self,
        objective_function: (
            Callable[
                [optuna.Trial, DataSchema, vaex.DataFrame],
                float | list[float] | tuple[float, ...],
            ]
            | Callable[
                [optuna.Trial, DataSchema, list[tuple[vaex.DataFrame, vaex.DataFrame]]],
                float | list[float] | tuple[float, ...],
            ]
        ),
        direction: OptimizeDirection | list[OptimizeDirection],
        num_trials: int,
        cv_folds: Callable[[DataSchema, vaex.DataFrame], list[tuple[vaex.DataFrame, vaex.DataFrame]]] | None = None,
        sampler: BaseSampler | None = None,
        pruner: optuna.pruners.BasePruner | None = None,
        study_name: str | None = None,
        storage: str | RDBStorage | None = None,
        load_if_exists: bool = False,
        random_state: int | None = 42,
        cache_directory: str | Path = "data/optuna-tuner",
        cache_size: int = 1,
    ) -> None:
        """Initializes a new OptunaTuner instance.

        For more information about Optuna, please refer to the documentation:
        https://optuna.readthedocs.io/en/stable/.

        Note:
            To visualize the optimization process, you can use the `optuna-dashboard`
            library. By specifying the `storage` parameter, the tuner will save
            the study to the specified file or storage.

            If a sqlite3 file path is defined, the optimization can be visualized using the
            `optuna-dashboard` command:
            ```bash
            optuna-dashboard sqlite:///PATH_TO_YOUR_OPTUNA_STORAGE.sqlite3
            ```

            The `study_name` parameter can be used to specify the name of the study,
            which will be displayed in the `optuna-dashboard` interface. If the
            `study_name` is not specified, the current date and time will be used.
            It is also referred to as the `study_id` in Optuna.

        Warning:
            The caching functionality of the objective function is implemented by
            serializing the function source code itself. Ensure that all dependencies
            of the objective function are defined within the function itself. Otherwise,
            the dependencies will not be included in the fingerprint of the tuner and
            the results of the hyperparameter tuning will be unpredictable. For example,
            if the objective function depends on a global variable, the caching
            functionality will not detect changes to the value itself and will not
            recompute the result.

            In addition, the objective function should preferably not use any cached
            methods, such as `BaseModel.fit_transform`. Instead, the objective
            function should use the underscored methods (`BaseModel._fit_transform`)
            to avoid caching the results of each trial.

        Args:
            objective_function: The objective function to optimize. The function must
                accept three arguments: the Optuna trial, the data schema, and the
                DataFrame or CV list of DataFrames to be tuned on. The function must
                return either a single float value or a list/tuple of float values.
                If a list/tuple is returned, the tuner will perform multi-objective
                optimization.
            direction: The direction of optimization. Either "maximize" or "minimize".
                If a list of directions is given, the tuner will perform multi-objective
                optimization. The length of the list must match the length of the list
                returned by the objective function.
            cv_folds: The cross-validation function to use. The function must accept
                the data schema and the DataFrame to be tuned on and return a list of
                tuples containing the training and validation DataFrames. The length
                of the list must match the number of folds to perform.
            num_trials: The number of trials to perform.
            sampler: The Optuna sampler to use, if None `TPESampler` is
                used for single-objective optimization and `NSGAIISampler`
                is used for multi-objective optimization.
            pruner: The Optuna pruner to use, if None `optuna.pruners.MedianPruner` is
                used.
            study_name: The name of the study. If None, the current date and time will
                be used.
            storage: The name of the storage to save the study to. If None, the
                study will not be saved to a persistent storage. Refer to the Optuna
                documentation for more information on the storage options.
            load_if_exists: Flag to control the behavior to handle a conflict of study
                names. In the case where a study named `study_name` already exists in
                the storage, a `DuplicatedStudyError` is raised if `load_if_exists`
                is set to `False`. Otherwise, the creation of the study is
                skipped, and the existing one is returned.
            random_state: The random state to use for the Optuna sampler. If None, the
                default random state of the sampler is used. Setting this will override
                the random state of the sampler.
            cache_directory: The target directory where the output is to be saved.
            cache_size: The maximum number of cache entries.

        Examples:
            >>> import vaex
            >>> from mleko.model import LGBMModel
            >>> from mleko.tune import OptunaTuner
            >>> from mleko.dataset import DataSchema
            >>> def objective_function(trial, data_schema, dataframe):
            ...     params = {
            ...         "num_iterations": trial.suggest_int("num_iterations", 10, 100),
            ...         "learning_rate": trial.suggest_float("learning_rate", 0.01, 0.1),
            ...         "num_leaves": trial.suggest_int("num_leaves", 10, 100),
            ...     }
            >>>     model = LGBMModel(
            ...         target="class_",
            ...         features=["sepal_width", "petal_length", "petal_width"],
            ...         num_iterations=100,
            ...         learning_rate=0.1,
            ...         num_leaves=31,
            ...         random_state=42,
            ...         metric=["auc"],
            ...     )
            >>>     df_train, df_val = dataframe.ml.random_split(test_size=0.20, verbose=False)
            >>>     _, metrics, _, _ = model._fit_transform(data_schema, df_train, df_val, params)
            >>>     return metrics['validation']['auc'][-1]
            >>> optuna_tuner = OptunaTuner(
            ...     objective_function=objective_function,
            ...     direction="maximize",
            ...     num_trials=51,
            ...     random_state=42,
            ... )
            >>> dataframe = vaex.ml.datasets.load_iris()
            >>> data_schema = DataSchema(
            ...     numerical=["sepal_length", "sepal_width", "petal_length", "petal_width"],
            ... )
            >>> best_trial, best_score, study = optuna_tuner.tune(data_schema, dataframe)
        """
        super().__init__(cache_directory, cache_size)
        self._objective_function = objective_function
        self._direction = direction
        self._num_trials = num_trials
        self._cv_folds = cv_folds
        self._sampler = sampler or (TPESampler() if isinstance(direction, list) else NSGAIISampler())
        self._pruner = pruner or optuna.pruners.MedianPruner()
        self._study_name = study_name if study_name is not None else datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self._storage = storage
        self._load_if_exists = load_if_exists
        self._using_optuna_dashboard = True if self._storage is not None else False
        self._random_state = random_state

        self._reset_sampler_rng(self._sampler)

        optuna_logger = logging.getLogger("optuna")
        for handler in optuna_logger.handlers:
            optuna_logger.removeHandler(handler)
        for handler in logger.handlers:
            optuna_logger.addHandler(handler)

        if self._storage is not None:
            storage_name = self._storage if isinstance(self._storage, str) else self._storage.url
            logger.info(
                f"Optuna study named {self._study_name!r} is stored in {storage_name}, use optuna-dashboard to view it."
            )

    def _tune(
        self, data_schema: DataSchema, dataframe: vaex.DataFrame
    ) -> tuple[HyperparametersType, float | list[float] | tuple[float, ...], optuna.study.Study]:
        """Perform the hyperparameter tuning.

        Args:
            data_schema: Data schema for the DataFrame.
            dataframe: DataFrame to be tuned on.

        Returns:
            Tuple containing the best hyperparameters, the best score, and a the Optuna study.
        """
        self._reset_sampler_rng(self._sampler)
        study = optuna.create_study(
            storage=self._storage,
            sampler=self._sampler,
            pruner=self._pruner,
            study_name=self._study_name,
            direction=self._direction if isinstance(self._direction, str) else None,
            directions=self._direction if isinstance(self._direction, list) else None,
            load_if_exists=self._load_if_exists,
        )
        if self._using_optuna_dashboard:
            direction_str = (
                self._direction if isinstance(self._direction, str) else ", ".join(self._direction)
            ).upper()
            save_note(
                study,
                f"""# Experiment Documentation

## Overview
| | |
|---|---|
| **Study Name** | {self._study_name} |
| **Load If Exists** | `{self._load_if_exists}` |
| **Creation Date** | {datetime.now().strftime("%Y-%m-%d %H:%M:%S")} |
| **Creator** | {getpass.getuser()} |
---

## Methodology

### Experimental Setup

| | |
|---|---|
| **Python Version** | `{platform.python_version()}` |
| **Optuna Version** | `{optuna.__version__}` |
| **Mleko Version** | `{mleko_version}` |
| **Random State** | `{self._random_state}` |
| **Sampler** | `{self._sampler.__class__.__name__}` |
| **Pruner** | `{self._pruner.__class__.__name__}` |
| **Optimization Direction** | {direction_str} |

## Notes
> _Add any additional notes about the experiment here._

## Appendix
### Objective Function
```python
{inspect.getsource(self._objective_function)}
```
{"### Cross-Validation Folds Function" if self._cv_folds is not None else ""}
{"```python" if self._cv_folds is not None else ""}
{inspect.getsource(self._cv_folds) if self._cv_folds is not None else ""}
{"```" if self._cv_folds is not None else ""}
                """,
            )

        with tqdm(total=self._num_trials) as pbar:

            def tqdm_callback(study: optuna.study.Study, _trial: optuna.trial.FrozenTrial) -> None:
                pbar.update(1)
                best_trial = study.best_trials[0].number
                best_score = study.best_trials[0].values

                if isinstance(best_score, (list, tuple)) and len(best_score) == 1:
                    best_values = f"{best_score[0]:.4f}"
                else:
                    best_values = f"[{', '.join(f'{x:.4f}' for x in best_score)}]"

                pbar.set_description(f"Best trial: {best_trial} | Best score: {best_values}")

            if self._cv_folds is not None:
                cv_folds = [
                    (df_train.copy(), df_val.copy()) for df_train, df_val in self._cv_folds(data_schema, dataframe)
                ]
                study.optimize(
                    lambda trial: self._objective_function(trial, data_schema.copy(), cv_folds),  # type: ignore
                    n_trials=self._num_trials,
                    callbacks=[tqdm_callback],
                )
            else:
                study.optimize(
                    lambda trial: self._objective_function(trial, data_schema.copy(), dataframe.copy()),
                    n_trials=self._num_trials,
                    callbacks=[tqdm_callback],
                )

        best_parameters = study.best_trials[0].params
        best_score = study.best_trials[0].values
        best_score = best_score[0] if isinstance(best_score, (list, tuple)) and len(best_score) == 1 else best_score

        self._reset_sampler_rng(self._sampler)

        return best_parameters, best_score, study

    def _fingerprint(self) -> Hashable:
        """Generates a fingerprint for the tuner.

        Returns:
            The fingerprint of the tuner.
        """
        if self._load_if_exists and self._storage is not None:  # pragma: no cover
            logger.warning("The load_if_exists parameter is set to True, the caching functionality will be disabled.")
            return random.random()

        return (
            super()._fingerprint(),
            CallableSourceFingerprinter().fingerprint(self._objective_function),
            self._direction,
            self._num_trials,
            CallableSourceFingerprinter().fingerprint(self._cv_folds) if self._cv_folds is not None else None,
            OptunaSamplerFingerprinter().fingerprint(self._sampler),
            OptunaPrunerFingerprinter().fingerprint(self._pruner),
            self._random_state,
            (
                self._storage
                if isinstance(self._storage, str)
                else (
                    self._storage.url + DictFingerprinter().fingerprint(self._storage.engine_kwargs)
                    if self._storage is not None
                    else None
                )
            ),
            self._load_if_exists,
        )

    def _reset_sampler_rng(self, sampler: BaseSampler) -> None:
        """Resets the random number generator of the given Optuna sampler.

        Args:
            sampler: The Optuna sampler to reset the random number generator of.
        """
        if isinstance(
            sampler,
            (
                GridSampler,
                RandomSampler,
                TPESampler,
                NSGAIISampler,
                NSGAIIISampler,
                BruteForceSampler,
            ),
        ):
            sampler._rng.rng.seed(self._random_state)

        if isinstance(
            sampler,
            (
                TPESampler,
                NSGAIISampler,
                NSGAIIISampler,
            ),
        ):
            self._reset_sampler_rng(sampler._random_sampler)

        if isinstance(sampler, CmaEsSampler):
            sampler._cma_rng.rng.seed(self._random_state)

        if isinstance(sampler, (CmaEsSampler, QMCSampler)):
            self._reset_sampler_rng(sampler._independent_sampler)

        if isinstance(sampler, PartialFixedSampler):
            self._reset_sampler_rng(sampler._base_sampler)

        if isinstance(sampler, (NSGAIISampler, NSGAIIISampler)) and isinstance(
            sampler._child_generation_strategy, NSGAIIChildGenerationStrategy
        ):
            sampler._child_generation_strategy._rng.rng.seed(self._random_state)

        if isinstance(sampler, QMCSampler) and self._random_state is not None:
            sampler._seed = self._random_state



"""This package provides Fingerprinter classes for generating unique fingerprints.

Fingerprinters are used for generating unique fingerprints of various data and file types,
such as Vaex DataFrames or CSV files. These fingerprints can be used to track changes in data and support
caching mechanisms.

The following fingerprinting utilities are provided:
    - `BaseFingerprinter`: The base class for all fingerprinters.
    - `CSVFingerprinter`: A fingerprinter for CSV files.
    - `VaexFingerprinter`: A fingerprinter for Vaex DataFrames.
    - `DictFingerprinter`: A fingerprinter for dictionaries.
    - `CallableSourceFingerprinter`: A fingerprinter for Python Callables that hashes the source code of the Callable
        to generate a fingerprint.
    - `OptunaSamplerFingerprinter`: A fingerprinter for Optuna samplers.
    - `OptunaPrunerFingerprinter`: A fingerprinter for Optuna pruners.
"""

from __future__ import annotations

from .base_fingerprinter import BaseFingerprinter
from .callable_source_fingerprinter import CallableSourceFingerprinter
from .csv_fingerprinter import CSVFingerprinter
from .dict_fingerprinter import DictFingerprinter
from .optuna_pruner_fingerprinter import OptunaPrunerFingerprinter
from .optuna_sampler_fingerprinter import OptunaSamplerFingerprinter
from .vaex_fingerprinter import VaexFingerprinter


__all__ = [
    "BaseFingerprinter",
    "CallableSourceFingerprinter",
    "CSVFingerprinter",
    "VaexFingerprinter",
    "OptunaPrunerFingerprinter",
    "OptunaSamplerFingerprinter",
    "DictFingerprinter",
]



"""This module contains the abstract base class for creating specialized fingerprinters.

The fingerprinter is used to generate a unique identifier for the given data, which is used
to detect changes in the data. The fingerprinter is used by the cache to determine whether
the data has changed since the last time it was cached.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any


class BaseFingerprinter(ABC):
    """Abstract base class for creating specialized fingerprinters."""

    @abstractmethod
    def fingerprint(self, data: Any) -> str:
        """Generate a fingerprint for the given data.

        The fingerprint should be a unique identifier for the given data, across different
        runs of the program, i.e. the fingerprint should be the same for the same data
        regardless of when the program is run.

        Args:
            data: Data that should be fingerprinted.

        Raises:
            NotImplementedError: The method has to be implemented by the subclass.

        Returns:
            str: The fingerprint as a hexadecimal string.
        """
        raise NotImplementedError



"""The subpackage provides functionality for transforming features.

This subpackage offers a collection of feature transformers, each designed for a specific type of feature
transformation task. By using these feature transformers sequentially, you can create a complete feature
transformation workflow within the pipeline.

The following feature transformers are provided by the subpackage:
    - `BaseTransformer`: The abstract base class for all feature transformers.
    - `CompositeTransformer`: A feature transformer that combines multiple feature transformers into a single feature
        transformer.
    - `FrequencyEncoderTransformer`: A feature transformer for encoding categorical features using frequency encoding.
    - `LabelEncoderTransformer`: A feature transformer for encoding categorical features using label encoding.
    - `MaxAbsScalerTransformer`: A feature transformer for scaling features using maximum absolute scaling.
    - `MinMaxScalerTransformer`: A feature transformer for scaling features using min-max scaling.
"""

from .base_transformer import BaseTransformer
from .composite_transformer import CompositeTransformer
from .frequency_encoder_transformer import FrequencyEncoderTransformer
from .label_encoder_transformer import LabelEncoderTransformer
from .max_abs_scaler_transformer import MaxAbsScalerTransformer
from .min_max_scaler_transformer import MinMaxScalerTransformer


__all__ = [
    "BaseTransformer",
    "CompositeTransformer",
    "FrequencyEncoderTransformer",
    "LabelEncoderTransformer",
    "MaxAbsScalerTransformer",
    "MinMaxScalerTransformer",
]



"""Module for the label encoder transformer."""

from __future__ import annotations

import re
from pathlib import Path
from typing import Hashable

import vaex
import vaex.array_types

from mleko.cache.fingerprinters import DictFingerprinter
from mleko.dataset.data_schema import DataSchema
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr
from mleko.utils.vaex_helpers import get_column

from .base_transformer import BaseTransformer


logger = CustomLogger()
"""A module-level logger for the module."""


class LabelEncoderTransformer(BaseTransformer):
    """Transforms features using label encoding."""

    @auto_repr
    def __init__(
        self,
        features: list[str] | tuple[str, ...],
        label_dict: dict[str, dict[str | None, int | None]] | None = None,
        allow_unseen: bool = False,
        encode_null: bool = False,
        cache_directory: str | Path = "data/label-encoder-transformer",
        cache_size: int = 1,
    ) -> None:
        """Initializes the transformer.

        Encodes categorical features with integer values between `0` and `n_classes-1`. If a value is not seen during
        fitting, it will be encoded as `-2`, unless `allow_unseen` is set to False, in which case an error will be
        raised. If `encode_null` is set to True, null values will be encoded as `-1`, otherwise they will be kept as
        `None`.

        Warning:
            Should only be used with categorical features of string type.

        Note:
            If `label_dict` is not provided, during fitting, the transformer will assign label mappings from the data
            with no guarantee of consistency across different runs. If `label_dict` is provided, keep in mind that
            the mappings must be integers between `0` and `n_classes-1` or `None`. Otherwise, the transformer will
            raise an error during fitting.

        Args:
            features: List of feature names to be used by the transformer.
            label_dict: A dictionary of label mappings dicts for each feature. Encoded labels must be integers
                between `0` and `n_classes-1` or `None`. If only some features are provided, the transformer will
                only use the provided label mappings for those features, while the rest will be assigned during
                fitting. If not provided at all, the transformer will assign label mappings for all features
                during fitting.
            allow_unseen: Whether to allow unseen values once the transformer is fitted.
            encode_null: Whether to encode null values as a separate category or keep them as null.
            cache_directory: Directory where the cache will be stored locally.
            cache_size: The maximum number of entries to keep in the cache.

        Examples:
            >>> import vaex
            >>> from mleko.dataset.transform import LabelEncoderTransformer
            >>> df = vaex.from_arrays(
            ...     a=["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"],
            ...     b=["a", "a", "a", "a", None, None, None, None, None, None],
            ...     c=["a", "b", "b", "b", "b", "b", None, None, None, None],
            ... )
            >>> ds = DataSchema(
            ...     categorical=["a", "b", "c"],
            ... )
            >>> _, _, df = LabelEncoderTransformer(
            ...     features=["a", "b"],
            ...     allow_unseen=True,
            ...     label_dict={  # Optional, but recommended
            ...         "a": {
            ...             "a": 0,
            ...             "b": 1,
            ...             "c": 2,
            ...             "d": 3,
            ...             "e": 4,
            ...             "f": 5,
            ...             "g": 6,
            ...             "h": 7,
            ...             "i": 8,
            ...             "j": 9,
            ...         },
            ...         "b": {
            ...             "a": 1,
            ...             "b": 0,
            ...         },
            ...         "c": {
            ...             "a": 1,
            ...             "b": 0,
            ...         },
            ...     },
            ... ).fit_transform(ds, df)
            >>> df["a"].tolist()
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            >>> df["b"].tolist()
            [1, 1, 1, 1, None, None, None, None, None, None]
            >>> df["c"].tolist()
            [1, 0, 0, 0, 0, 0, None, None, None, None]
        """
        super().__init__(features, cache_directory, cache_size)
        self._allow_unseen = allow_unseen
        self._encode_null = encode_null
        self._label_dict = label_dict

        if label_dict is None:
            if self._encode_null:
                logger.warning("Null values will be encoded as `-1`.")
            else:
                logger.warning("Null values will be kept as `None`.")

        self._transformer = {}
        if label_dict is not None:
            missing_features = [feature for feature in features if feature not in label_dict]
            if missing_features:
                msg = (
                    f"Label dictionary is missing some features and will automatically assign "
                    f"mappings during fitting for the following features: {missing_features}."
                )
                logger.warning(msg)

            self._transformer = label_dict

    def _fit(
        self, data_schema: DataSchema, dataframe: vaex.DataFrame
    ) -> tuple[DataSchema, dict[str, dict[str | None, int | None]] | dict[str, dict[str | None, int]]]:
        """Fits the transformer on the given DataFrame.

        Args:
            data_schema: The data schema of the DataFrame.
            dataframe: The DataFrame to fit the transformer on.

        Returns:
            Updated data schema and fitted transformer.
        """
        logger.info(f"Fitting label encoder transformer ({len(self._features)}): {self._features}.")
        for feature in self._features:
            self._ensure_valid_feature_type(feature, data_schema, dataframe)
            labels: list[str] = get_column(dataframe, feature).unique(dropna=True)  # type: ignore
            if not self._fit_using_label_dict(feature, labels):
                logger.info(f"Assigning mappings for feature {feature!r}: {labels}.")
                self._transformer[feature] = {label: i for i, label in enumerate(labels)}
                self._transformer[feature][None] = -1 if self._encode_null else None

        return data_schema, self._transformer

    def _transform(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> tuple[DataSchema, vaex.DataFrame]:
        """Transforms the features of the given DataFrame using label encoding.

        Args:
            data_schema: The data schema of the DataFrame.
            dataframe: The DataFrame to transform.

        Returns:
            Updated data schema and transformed DataFrame.
        """
        logger.info(f"Transforming features using label encoding ({len(self._features)}): {self._features}.")
        df = dataframe.copy()
        for feature in self._features:
            try:
                df[feature] = get_column(df, feature).map(
                    self._transformer[feature],
                    default_value=-2 if self._allow_unseen else None,
                )
            except ValueError as e:
                if not self._allow_unseen and re.search(r"Missing \d values in mapper", str(e)):
                    msg = (
                        f"Unseen values encountered during transformation for feature {feature!r}. "
                        "Set `allow_unseen` to True to convert unseen values to -1 instead of raising an error."
                    )
                    logger.error(msg)
                    raise ValueError(msg) from e
                raise e  # pragma: no cover

        return data_schema, df

    def _fingerprint(self) -> Hashable:
        """Returns the fingerprint of the transformer.

        Appends the `allow_unseen`, `encode_null`, and `label_dict` attributes to the fingerprint.

        Returns:
            A hashable object that uniquely identifies the transformer.
        """
        return (
            super()._fingerprint(),
            self._allow_unseen,
            self._encode_null,
            DictFingerprinter().fingerprint(self._label_dict),
        )

    def _fit_using_label_dict(self, feature: str, observed_labels: list[str]) -> bool:
        """Attempts to fit the label dictionary for the specified feature.

        If the label dictionary is not provided or the feature is not in the label dictionary, the function will
        return False. Otherwise, it will fit the label dictionary and return True.

        Args:
            feature: The feature to fit the label dictionary for.
            observed_labels: The observed labels for the feature.

        Raises:
            ValueError: If the label dictionary contains invalid mappings.

        Returns:
            Whether the label dictionary was fitted.
        """
        if self._label_dict is None or feature not in self._label_dict:
            return False

        self._transformer[feature] = self._label_dict[feature]
        if any(encoding is not None and encoding < 0 for encoding in self._transformer[feature].values()):
            msg = (
                f"Label dictionary for feature {feature!r} must contain non-negative integers or None. "
                f"Found: {self._transformer[feature]}."
            )
            logger.error(msg)
            raise ValueError(msg)

        missing_labels = set(observed_labels) - set(self._transformer[feature].keys())
        if missing_labels:
            logger.warning(
                f"Label dictionary for feature {feature!r} is missing labels: {missing_labels}. "
                "Assigning mappings during fitting."
            )
            max_label = max((value for value in self._transformer[feature].values() if value is not None), default=-1)
            self._transformer[feature].update({label: i for i, label in enumerate(missing_labels, start=max_label + 1)})

        null_mapping = self._transformer[feature].get(None, -1 if self._encode_null else None)
        if isinstance(null_mapping, int) and null_mapping != -1 and self._encode_null:
            logger.warning(f"Null values for feature {feature!r} will be encoded as `-1`, not {null_mapping}.")
            null_mapping = -1
        self._transformer[feature][None] = null_mapping
        return True

    def _ensure_valid_feature_type(self, feature: str, data_schema: DataSchema, dataframe: vaex.DataFrame) -> None:
        """Check if the feature is of the correct type.

        Args:
            feature: The feature to check.
            data_schema: The data schema of the DataFrame.
            dataframe: The DataFrame to check.

        Raises:
            ValueError: If the feature is not of the correct type.
        """
        feature_type = data_schema.get_type(feature)
        feature_dtype = get_column(dataframe, feature).dtype
        if feature_type != "categorical" or feature_dtype != "string":
            msg = (
                f"LabelEncoderTransformer can only be used with categorical features of string type. "
                f"Feature {feature!r} is not (type: {feature_type}, dtype: {feature_dtype})."
            )
            logger.error(msg)
            raise ValueError(msg)



"""Module containing the FilterStep class."""

from __future__ import annotations

from typing import Union

from typing_extensions import TypedDict
from vaex import DataFrame

from mleko.dataset.data_schema import DataSchema
from mleko.dataset.filter.base_filter import BaseFilter
from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import PipelineStep
from mleko.utils.decorators import auto_repr


class FilterStepInputType(TypedDict):
    """The input type of the FilterStep."""

    data_schema: Union[str, DataSchema]
    """The key for the DataSchema."""

    dataframe: Union[str, DataFrame]
    """The key for the DataFrame to be filtered."""


class FilterStepOutputType(TypedDict):
    """The output type of the FilterStep."""

    dataframe: str
    """The key for the filtered DataFrame."""


class FilterStep(PipelineStep):
    """Pipeline step that filters a DataFrame."""

    _inputs: FilterStepInputType
    _outputs: FilterStepOutputType

    @auto_repr
    def __init__(
        self,
        filter: BaseFilter,
        inputs: FilterStepInputType,
        outputs: FilterStepOutputType,
        cache_group: str | None = None,
    ) -> None:
        """Initialize the FilterStep with the specified data filter.

        Args:
            filter: The DataFilter responsible for handling data filtering.
            inputs: A dictionary of input keys following the `FilterStepInputType` schema.
            outputs: A dictionary of output keys following the `FilterStepOutputType` schema.
            cache_group: The cache group to use.
        """
        super().__init__(inputs, outputs, cache_group)
        self._filter = filter

    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Perform data filtering using the configured filter.

        Args:
            data_container: Contains the DataFrame to be filtered.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Returns:
            A DataContainer containing the result.
        """
        data_schema = self._validate_and_get_input(self._inputs["data_schema"], DataSchema, data_container)
        dataframe = self._validate_and_get_input(self._inputs["dataframe"], DataFrame, data_container)

        filtered_dataframe = self._filter.filter(data_schema, dataframe, self._cache_group, force_recompute)
        data_container.data[self._outputs["dataframe"]] = filtered_dataframe
        return data_container

    def _get_input_model(self) -> type[FilterStepInputType]:
        """Get the input type for the FilterStep.

        Returns:
            Input type for the FilterStep.
        """
        return FilterStepInputType

    def _get_output_model(self) -> type[FilterStepOutputType]:
        """Get the output type for the FilterStep.

        Returns:
            Output type for the FilterStep.
        """
        return FilterStepOutputType



"""Module containing the ModelStep class."""

from __future__ import annotations

from typing import Optional, Union, cast

from typing_extensions import TypedDict
from vaex import DataFrame

from mleko.dataset.data_schema import DataSchema
from mleko.model.base_model import BaseModel, HyperparametersType
from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import FitTransformAction, FitTransformPipelineStep
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr


logger = CustomLogger()
"""The logger for the module."""


class ModelStepInputFitType(TypedDict):
    """The input type of the ModelStep when action is 'fit'."""

    data_schema: Union[str, DataSchema]
    """DataSchema or the key for the DataSchema to be used for training."""

    dataframe: Union[str, DataFrame]
    """DataFrame or the key for the DataFrame to be used for training."""

    validation_dataframe: Optional[Union[str, DataFrame]]
    """DataFrame or the key for the validation DataFrame to be used for training."""

    hyperparameters: Optional[Union[str, HyperparametersType]]
    """Hyperparameters or the key for the hyperparameters to be used for training."""


class ModelStepInputTransformType(TypedDict):
    """The input type of the ModelStep when action is 'transform'."""

    data_schema: Union[str, DataSchema]
    """DataSchema or the key for the DataSchema to be used for prediction."""

    dataframe: Union[str, DataFrame]
    """DataFrame or the key for the DataFrame to be used for prediction."""


class ModelStepOutputFitType(TypedDict):
    """The output type of the ModelStep when action is 'fit'."""

    model: str
    """The key for the fitted model after training."""

    metrics: str
    """The key for the metrics dictionary after training."""


class ModelStepOutputTransformType(TypedDict):
    """The output type of the ModelStep when action is 'transform'."""

    dataframe: str
    """The key for the transformed DataFrame after prediction."""


class ModelStepOutputFitTransformType(ModelStepOutputFitType, ModelStepOutputTransformType):
    """The output type of the ModelStep when action is 'fit_transform'."""

    validation_dataframe: Optional[str]
    """The key for the transformed validation DataFrame after prediction."""


class ModelStep(FitTransformPipelineStep):
    """Pipeline step for model training and prediction."""

    _inputs: ModelStepInputFitType | ModelStepInputTransformType
    _outputs: ModelStepOutputFitType | ModelStepOutputTransformType | ModelStepOutputFitTransformType

    @auto_repr
    def __init__(
        self,
        model: BaseModel,
        action: FitTransformAction,
        inputs: ModelStepInputFitType | ModelStepInputTransformType,
        outputs: ModelStepOutputFitType | ModelStepOutputTransformType | ModelStepOutputFitTransformType,
        cache_group: str | None = None,
    ) -> None:
        """Initialize the ModelStep with the specified model.

        The action parameter specifies whether the model should be fitted, transformed, or both. If the action
        is "fit" the `ModelStep` will return the fitted model. If the action is "transform" the
        `ModelStep` will return the predictions of the DataFrame. If the action is "fit_transform" the `ModelStep`
        will return both the fitted model and the predictions of the DataFrame in the specified order.

        Note:
            The hyperparameters input is optional. If it is not provided, the model will be trained using the
            default hyperparameters configured in the model.

        Args:
            model: The model used for training and prediction.
            action: The action to perform, one of "fit", "transform", or "fit_transform".
            inputs: A dictionary of input keys following the `ModelStepInputFitType` or `ModelStepInputTransformType`
                schema, depending on the action.
            outputs: A list of output keys following the `ModelStepOutputFitType`, `ModelStepOutputTransformType`, or
                `ModelStepOutputFitTransformType` schema, depending on the action.
            cache_group: The cache group to use.
        """
        super().__init__(action, inputs, outputs, cache_group)
        self._model = model

    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Perform actions using the configured model.

        Args:
            data_container: Contains the input.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Returns:
            A DataContainer containing the output of the action performed by the step, either the fitted model,
            the predictions on the DataFrame, or both.
        """
        data_schema = self._validate_and_get_input(self._inputs["data_schema"], DataSchema, data_container)
        dataframe = self._validate_and_get_input(self._inputs["dataframe"], DataFrame, data_container)

        df_validation = hyperparameters = None
        if self._action != "transform":
            self._inputs = cast(ModelStepInputFitType, self._inputs)
            df_validation = self._validate_and_get_input(
                self._inputs["validation_dataframe"], DataFrame, data_container, is_optional=True
            )
            hyperparameters: HyperparametersType | None = self._validate_and_get_input(
                self._inputs["hyperparameters"], dict, data_container, is_optional=True
            )

        if self._action == "fit":
            self._outputs = cast(ModelStepOutputFitType, self._outputs)
            model, metrics = self._model.fit(
                data_schema, dataframe, df_validation, hyperparameters, self._cache_group, force_recompute
            )
            data_container.data[self._outputs["model"]] = model
            data_container.data[self._outputs["metrics"]] = metrics
        elif self._action == "transform":
            self._outputs = cast(ModelStepOutputTransformType, self._outputs)
            df = self._model.transform(data_schema, dataframe, self._cache_group, force_recompute)
            data_container.data[self._outputs["dataframe"]] = df
        elif self._action == "fit_transform":
            self._outputs = cast(ModelStepOutputFitTransformType, self._outputs)
            model, metrics, df, df_validation = self._model.fit_transform(
                data_schema, dataframe, df_validation, hyperparameters, self._cache_group, force_recompute
            )
            data_container.data[self._outputs["model"]] = model
            data_container.data[self._outputs["metrics"]] = metrics
            data_container.data[self._outputs["dataframe"]] = df
            if self._outputs["validation_dataframe"] is not None:
                data_container.data[self._outputs["validation_dataframe"]] = df_validation
        return data_container

    def _get_input_model(self) -> type[ModelStepInputFitType | ModelStepInputTransformType]:
        """Get the input type for the TransformStep.

        Returns:
            Input type for the TransformStep.
        """
        if self._action == "fit" or self._action == "fit_transform":
            return ModelStepInputFitType
        return ModelStepInputTransformType

    def _get_output_model(
        self,
    ) -> type[ModelStepOutputFitType | ModelStepOutputTransformType | ModelStepOutputFitTransformType]:
        """Get the output type for the TransformStep.

        Returns:
            Output type for the TransformStep.
        """
        if self._action == "fit":
            return ModelStepOutputFitType
        if self._action == "transform":
            return ModelStepOutputTransformType
        return ModelStepOutputFitTransformType



"""The module contains a fingerprinter for CSV files supporting Gzipped and raw CSV files."""

from __future__ import annotations

import gzip
import hashlib
from concurrent import futures
from itertools import islice
from pathlib import Path

from mleko.utils.custom_logger import CustomLogger

from .base_fingerprinter import BaseFingerprinter


logger = CustomLogger()
"""The logger for the module."""


class CSVFingerprinter(BaseFingerprinter):
    """A fingerprinter for CSV files supporting Gzipped and raw CSV files."""

    def __init__(self, n_rows: int = 1000):
        """Initialize the CSVFingerprinter.

        Warning:
            The fingerprint is generated by reading the first `n_rows` of each CSV file. If the
            CSV file is larger than `n_rows`, only the first `n_rows` are read. This means that
            the fingerprint is not unique for the entire CSV file, but only for the first `n_rows`.

        Args:
            n_rows: The number of rows to sample from each CSV file for fingerprinting.

        Examples:
            >>> fingerprinter = CSVFingerprinter(n_rows=1000)
            >>> fingerprinter.fingerprint(["data.csv", "data2.csv"])
            "fingerprint"
        """
        self._n_rows = n_rows

    def fingerprint(self, data: list[str] | list[Path]) -> str:
        """Generate a fingerprint for the given list of CSV files.

        The currently supported file types are `.csv`, `.gz`, and `.csv.gz`.

        Args:
            data: A list of file paths to CSV files.

        Returns:
            The fingerprint as a hexadecimal string.
        """
        file_posix_paths: list[Path] = [Path(file_path) for file_path in data]
        with futures.ThreadPoolExecutor(max_workers=None) as executor:
            file_fingerprints = list(executor.map(self._fingerprint_csv_file, file_posix_paths))

        file_fingerprints.sort()
        fingerprint = hashlib.md5("".join(file_fingerprints).encode()).hexdigest()
        return fingerprint

    def _fingerprint_csv_file(self, file_path: Path) -> str:
        """Generate a fingerprint for a single CSV file.

        Args:
            file_path: The file path to a CSV file.

        Raises:
            ValueError: File is unsupported file type.

        Returns:
            The fingerprint as a hexadecimal string.
        """
        if file_path.suffix not in {".csv", ".gz", ".csv.gz"}:
            msg = f"Unsupported file type: {file_path.suffix}"
            logger.error(msg)
            raise ValueError(msg)

        if file_path.suffix in {".gz", ".csv.gz"}:
            with gzip.open(file_path, "rb") as f:
                sample = b"".join(islice((f.readline() for _ in range(self._n_rows)), self._n_rows))
        else:
            with open(file_path, "rb") as f:
                sample = b"".join(islice((f.readline() for _ in range(self._n_rows)), self._n_rows))
        fingerprint = hashlib.md5(str(sample).encode()).hexdigest()
        return fingerprint



"""The subpackage provides functionality for converting data between different file formats.

The subpackage contains the following converter classes:
    - `BaseConverter`: The abstract base class for all converters.
    - `CSVToVaexConverter`: A converter for converting CSV files to Vaex DataFrames.
"""

from __future__ import annotations

from .base_converter import BaseConverter
from .csv_to_vaex_converter import CSVToVaexConverter


__all__ = ["BaseConverter", "CSVToVaexConverter"]



"""Custom logging module providing a CustomLogger and CustomFormatter.

This module defines CustomLogger and CustomFormatter classes that extend the functionality of the standard Python
logging module. CustomLogger provides colored output, configurable log levels, and improved stack trace information.
CustomFormatter enables colored log level names and custom formatting of log records.
"""

from __future__ import annotations

import logging
import re
import sys
from typing import Any, Mapping


class CustomFormatter(logging.Formatter):
    """A custom formatter to add colors and more detailed information for log messages."""

    RESET_SEQ = "\033[0m"
    COLOR_SEQ = "\033[1;%dm"
    BOLD_SEQ = "\033[1m"

    COLORS = {
        logging.INFO: COLOR_SEQ % 32,  # Green
        logging.WARNING: COLOR_SEQ % 33,  # Yellow
        logging.ERROR: COLOR_SEQ % 31,  # Red
        logging.CRITICAL: COLOR_SEQ % 35,  # Magenta
    }

    FORMAT = f"[%(asctime)s] [%(levelname)-8s] %(message)s {BOLD_SEQ}(%(filename)s:%(lineno)d){RESET_SEQ}"

    def __init__(self) -> None:
        """Initialize the custom formatter with the desired date and message format."""
        super().__init__(fmt=self.FORMAT, datefmt="%Y-%m-%d %H:%M:%S")

    def format(self, record: logging.LogRecord) -> str:
        """Format a log record according to the specified format.

        Args:
            record: The log record to be formatted.

        Returns:
            The formatted log message as a string.
        """
        level_color = self.COLORS.get(record.levelno, "")
        record.levelname = level_color + record.levelname + self.RESET_SEQ
        return super().format(record)


class CustomLogger(logging.Logger):
    """A custom logger class to use the `CustomFormatter` class."""

    _global_log_level: int = logging.INFO
    _instances: list[CustomLogger] = []

    def __init__(self) -> None:
        """Initialize the custom logger and set the log level and default handler."""
        super().__init__(__name__)
        self.setLevel(self._global_log_level)
        self._instances.append(self)

        if not self.handlers:
            handler = logging.StreamHandler(sys.stdout)
            handler.setFormatter(CustomFormatter())
            self.addHandler(handler)

    @staticmethod
    def set_global_log_level(log_level: int) -> None:
        """Set the global log level for all instances of CustomLogger.

        Args:
            log_level: The minimum log level to output.
        """
        CustomLogger._global_log_level = log_level
        for instance in CustomLogger._instances:
            instance.set_level(log_level)

    def debug(
        self,
        msg: object,
        *args: object,
        exc_info: Any = None,
        stack_info: bool = False,
        stacklevel: int = 2,
        extra: Mapping[str, object] | None = None,
    ) -> None:
        """Log a debug message.

        Args:
            msg: The message to be logged.
            args: Additional arguments propagated to the built-in `logging` module.
            exc_info: Exception information to log.
            stack_info: Whether to add stack information to the log message.
            stacklevel: Level in the stack frame to log the origin of this log record.
            extra: Additional information to log.
        """
        self._route_message(
            msg,
            logging.DEBUG,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )

    def info(
        self,
        msg: object,
        *args: object,
        exc_info: Any = None,
        stack_info: bool = False,
        stacklevel: int = 2,
        extra: Mapping[str, object] | None = None,
    ) -> None:
        """Log an info message.

        Args:
            msg: The message to be logged.
            args: Additional arguments propagated to the built-in `logging` module.
            exc_info: Exception information to log.
            stack_info: Whether to add stack information to the log message.
            stacklevel: Level in the stack frame to log the origin of this log record.
            extra: Additional information to log.
        """
        self._route_message(
            msg,
            logging.INFO,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )

    def warning(
        self,
        msg: object,
        *args: object,
        exc_info: Any = None,
        stack_info: bool = False,
        stacklevel: int = 2,
        extra: Mapping[str, object] | None = None,
    ) -> None:
        """Log a warning message.

        Args:
            msg: The message to be logged.
            args: Additional arguments propagated to the built-in `logging` module.
            exc_info: Exception information to log.
            stack_info: Whether to add stack information to the log message.
            stacklevel: Level in the stack frame to log the origin of this log record.
            extra: Additional information to log.
        """
        self._route_message(
            msg,
            logging.WARNING,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )

    def error(
        self,
        msg: object,
        *args: object,
        exc_info: Any = None,
        stack_info: bool = False,
        stacklevel: int = 2,
        extra: Mapping[str, object] | None = None,
    ) -> None:
        """Log an error message.

        Args:
            msg: The message to be logged.
            args: Additional arguments propagated to the built-in `logging` module.
            exc_info: Exception information to log.
            stack_info: Whether to add stack information to the log message.
            stacklevel: Level in the stack frame to log the origin of this log record.
            extra: Additional information to log.
        """
        self._route_message(
            msg,
            logging.ERROR,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )

    def critical(
        self,
        msg: object,
        *args: object,
        exc_info: Any = None,
        stack_info: bool = False,
        stacklevel: int = 2,
        extra: Mapping[str, object] | None = None,
    ) -> None:
        """Log a critical message.

        Args:
            msg: The message to be logged.
            args: Additional arguments propagated to the built-in `logging` module.
            exc_info: Exception information to log.
            stack_info: Whether to add stack information to the log message.
            stacklevel: Level in the stack frame to log the origin of this log record.
            extra: Additional information to log.
        """
        self._route_message(
            msg,
            logging.CRITICAL,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )

    def set_level(self, log_level: int) -> None:
        """Set the minimum log level for this CustomLogger instance.

        Args:
            log_level: The minimum logging level to output.
        """
        super().setLevel(log_level)

    def _route_message(
        self,
        message: object,
        log_level: int,
        *args: object,
        exc_info: Any,
        stack_info: bool,
        stacklevel: int,
        extra: Mapping[str, object] | None,
    ) -> None:
        """Route and clean the log message based on its content.

        Args:
            message: The log message to be routed.
            log_level: The original log level at which the message was sent.
            args: Additional arguments propagated to the built-in `logging` module.
            exc_info: Exception information to log.
            stack_info: Whether to add stack information to the log message.
            stacklevel: Level in the stack frame to log the origin of this log record.
            extra: Additional information to log.
        """
        target_stacklevel = stacklevel + 1
        if not isinstance(message, str):
            if log_level < self.level:  # pragma: no cover
                return

            self.log(
                log_level,
                message,
                *args,
                exc_info=exc_info,
                stack_info=stack_info,
                stacklevel=target_stacklevel,
                extra=extra,
            )
            return

        match = re.search(r"\[(Debug|Info|Warning|Error|Critical)\]", message, re.IGNORECASE)

        if match:
            level_str = match.group(1).upper()
            routed_level = getattr(logging, level_str)
            cleaned_msg = re.sub(rf"\[{level_str}\]\s?", "", message, flags=re.IGNORECASE, count=1)
            if routed_level < self.level:  # pragma: no cover
                return

            self.log(
                routed_level,
                cleaned_msg,
                *args,
                exc_info=exc_info,
                stack_info=stack_info,
                stacklevel=target_stacklevel,
                extra=extra,
            )
        else:
            if log_level < self.level:  # pragma: no cover
                return

            self.log(
                log_level,
                message,
                *args,
                exc_info=exc_info,
                stack_info=stack_info,
                stacklevel=target_stacklevel,
                extra=extra,
            )



"""The module contains a fingerprinter for `Optuna` pruners."""

from __future__ import annotations

import hashlib

import optuna

from mleko.utils.custom_logger import CustomLogger

from .base_fingerprinter import BaseFingerprinter


logger = CustomLogger()
"""The logger for the module."""


class OptunaPrunerFingerprinter(BaseFingerprinter):
    """Class to generate unique fingerprints for different types of Optuna pruners."""

    def fingerprint(self, data: optuna.pruners.BasePruner) -> str:
        """Generate a fingerprint string for a given Optuna pruner.

        Args:
            data: The pruner to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the pruner's configuration.
        """
        class_name = data.__class__.__name__.lower()
        method_name = f"_fingerprint_{class_name}"

        if hasattr(self, method_name):
            fingerprint_method = getattr(self, method_name)
            return fingerprint_method(data)
        else:
            logger.warning(f"Cannot fingerprint pruner of type {class_name}, ensure you invalidate the cache manually.")
            return class_name

    def _fingerprint_medianpruner(self, pruner: optuna.pruners.MedianPruner) -> str:
        """Generate a fingerprint string for an Optuna MedianPruner.

        Args:
            pruner: The MedianPruner to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the MedianPruner's configuration.
        """
        fingerprint = ""
        fingerprint += str(pruner._percentile)
        fingerprint += str(pruner._n_startup_trials)
        fingerprint += str(pruner._n_warmup_steps)
        fingerprint += str(pruner._interval_steps)
        fingerprint += str(pruner._n_min_trials)
        return hashlib.md5(fingerprint.encode()).hexdigest()

    def _fingerprint_noppruner(self, _pruner: optuna.pruners.NopPruner) -> str:
        """Generate a fingerprint string for an Optuna NopPruner.

        Args:
            pruner: The NopPruner to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the NopPruner's configuration.
        """
        fingerprint = ""
        return hashlib.md5(fingerprint.encode()).hexdigest()

    def _fingerprint_patientpruner(self, pruner: optuna.pruners.PatientPruner) -> str:
        """Generate a fingerprint string for an Optuna PatientPruner.

        Args:
            pruner: The PatientPruner to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the PatientPruner's configuration.
        """
        fingerprint = ""
        fingerprint += self.fingerprint(pruner._wrapped_pruner) if pruner._wrapped_pruner is not None else ""
        fingerprint += str(pruner._patience)
        fingerprint += str(pruner._min_delta)
        return hashlib.md5(fingerprint.encode()).hexdigest()

    def _fingerprint_percentilepruner(self, pruner: optuna.pruners.PercentilePruner) -> str:
        """Generate a fingerprint string for an Optuna PercentilePruner.

        Args:
            pruner: The PercentilePruner to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the PercentilePruner's configuration.
        """
        fingerprint = ""
        fingerprint += str(pruner._percentile)
        fingerprint += str(pruner._n_startup_trials)
        fingerprint += str(pruner._n_warmup_steps)
        fingerprint += str(pruner._interval_steps)
        fingerprint += str(pruner._n_min_trials)
        return hashlib.md5(fingerprint.encode()).hexdigest()

    def _fingerprint_successivehalvingpruner(self, pruner: optuna.pruners.SuccessiveHalvingPruner) -> str:
        """Generate a fingerprint string for an Optuna SuccessiveHalvingPruner.

        Args:
            pruner: The SuccessiveHalvingPruner to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the SuccessiveHalvingPruner's configuration.
        """
        fingerprint = ""
        fingerprint += str(pruner._min_resource)
        fingerprint += str(pruner._reduction_factor)
        fingerprint += str(pruner._min_early_stopping_rate)
        fingerprint += str(pruner._bootstrap_count)
        return hashlib.md5(fingerprint.encode()).hexdigest()

    def _fingerprint_hyperbandpruner(self, pruner: optuna.pruners.HyperbandPruner) -> str:
        """Generate a fingerprint string for an Optuna HyperbandPruner.

        Args:
            pruner: The HyperbandPruner to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the HyperbandPruner's configuration.
        """
        fingerprint = ""
        fingerprint += str(pruner._min_resource)
        fingerprint += str(pruner._max_resource)
        fingerprint += str(pruner._reduction_factor)
        fingerprint += str(pruner._bootstrap_count)
        return hashlib.md5(fingerprint.encode()).hexdigest()

    def _fingerprint_thresholdpruner(self, pruner: optuna.pruners.ThresholdPruner) -> str:
        """Generate a fingerprint string for an Optuna ThresholdPruner.

        Args:
            pruner: The ThresholdPruner to fingerprint.

        Returns:
            A fingerprint string that uniquely identifies the ThresholdPruner's configuration.
        """
        fingerprint = ""
        fingerprint += str(pruner._lower)
        fingerprint += str(pruner._upper)
        fingerprint += str(pruner._n_warmup_steps)
        fingerprint += str(pruner._interval_steps)
        return hashlib.md5(fingerprint.encode()).hexdigest()



"""Module for the LightGBM model."""

from __future__ import annotations

import logging
import pprint as pp
from pathlib import Path
from typing import Any, Hashable

import lightgbm as lgb
import numpy as np
import pandas as pd
import vaex
from lightgbm.sklearn import _LGBM_ScikitEvalMetricType

from mleko.dataset.data_schema import DataSchema
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr
from mleko.utils.vaex_helpers import HashableVaexDataFrame

from .base_model import BaseModel, HyperparametersType


logger = CustomLogger()
"""The logger for the module."""


def python_to_lgbm_verbosity(verbosity: int) -> int:  # pragma: no cover
    """Converts a Python `logging` level to a `LightGBM` verbosity level.

    Args:
        verbosity: The Python `logging` level (e.g., `logging.INFO`).

    Returns:
        The corresponding `LightGBM` verbosity level.
    """
    if verbosity <= logging.DEBUG:
        return 2
    elif verbosity == logging.INFO:
        return 1
    elif verbosity in (logging.WARNING, logging.ERROR):
        return 0
    elif verbosity >= logging.CRITICAL:
        return -1
    return 1


class LGBMModel(BaseModel):
    """Wrapper for the LightGBM model.

    Full documentation of the LightGBM model can be found here
    https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMModel.html.
    """

    @auto_repr
    def __init__(
        self,
        target: str,
        model: lgb.LGBMClassifier | lgb.LGBMRegressor,
        eval_metric: _LGBM_ScikitEvalMetricType | None = None,
        log_eval_period: int | None = 10,
        features: list[str] | tuple[str, ...] | None = None,
        ignore_features: list[str] | tuple[str, ...] | None = None,
        random_state: int | None = 42,
        verbosity: int = logging.INFO,
        memoized_dataset_cache_size: int | None = 0,
        cache_directory: str | Path = "data/lgbm-model",
        cache_size: int = 1,
    ) -> None:
        """Initialize the LightGBM model with the given hyperparameters.

        Note:
            Features and ignore_features are mutually exclusive. If both are provided, a `ValueError` will be raised.
            By default, all features are used. If ignore_features is provided, all features except the ones in
            ignore_features will be used. If features is provided, only the features in features will be used.

        Warning:
            The `memoized_dataset_cache_size` parameter is experimental and should be used with caution. It refers to
            the number of datasets to keep in memory for speeding up repeated training. This can be useful when
            hyperparameter tuning or cross-validation is performed, as the dataset does not need to be loaded from disk
            every time. However, this can lead to memory issues if the dataset is too large. Specify 0 to disable the
            cache. When finished with the fitting and transforming, please call the `_clear_dataset_cache` method to
            clear the cache and free up memory.

        Args:
            target: The name of the target feature.
            model: The LightGBM model to be used.
            eval_metric: Evaluation metric(s) to be used as list of strings or a single string. Refer to
                https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMClassifier.html#lightgbm.LGBMClassifier.fit
                to see the list of available metrics and how to define custom metrics.
            log_eval_period: The period to log the evaluation results.
            features: The names of the features to be used as input for the model.
            ignore_features: The names of the features to be ignored.
            random_state: The random state to be used for reproducibility.
            verbosity: The verbosity level of the logger, will be passed to the LightGBM model.
            memoized_dataset_cache_size: The number of datasets to keep in memory for speeding up repeated training.
                When finished with the fitting and transforming, please call the `_clear_dataset_cache` method to clear
                the cache and free up memory. Specify 0 to disable the cache.
            cache_directory: The target directory where the model will be saved.
            cache_size: The maximum number of entries to keep in the cache.

        Examples:
            >>> import vaex
            >>> from mleko.model import LGBMModel
            >>> from lightgbm import LGBMClassifier
            >>> df = vaex.ml.datasets.load_iris()
            >>> df_train, df_test = df.ml.random_split(test_size=0.20, verbose=False)
            >>> data_schema = DataSchema(
            ...    numerical=["sepal_length", "sepal_width", "petal_length", "petal_width"],
            ... )
            >>> model = LGBMModel(
            ...     target="class_",
            ...     model=LGBMClassifier(n_estimators=100),
            ...     random_state=42,
            ...     features=["sepal_width", "petal_length", "petal_width"],
            ... )
            >>> booster, df_train_pred, df_test_pred = model.fit_transform(data_schema, df_train, df_test, {})
        """
        super().__init__(features, ignore_features, verbosity, memoized_dataset_cache_size, cache_directory, cache_size)
        lgb.register_logger(logger)

        self._target = target
        self._model = model
        self._eval_metric = eval_metric
        self._log_eval_period = log_eval_period
        self._random_state = random_state

        self._model.set_params(
            random_state=self._random_state,
            verbosity=python_to_lgbm_verbosity(verbosity),
        )
        self._hyperparameters = self._model.get_params()
        logger.set_level(verbosity)

    def _fit(
        self,
        data_schema: DataSchema,
        train_dataframe: vaex.DataFrame,
        validation_dataframe: vaex.DataFrame | None = None,
        hyperparameters: HyperparametersType | None = None,
    ) -> tuple[lgb.LGBMClassifier | lgb.LGBMRegressor, dict[str, dict[str, list[Any]]]]:
        """Fits the LightGBM model to the given data with the given hyperparameters.

        Args:
            data_schema: The data schema of the dataframes.
            train_dataframe: The training dataframe.
            validation_dataframe: The validation dataframe, optional but required for early stopping.
            hyperparameters: The hyperparameters to use for training.

        Raises:
            ValueError: If the target feature is in the feature set.

        Returns:
            The trained LightGBM model.
        """
        self._model.set_params(random_state=self._random_state)
        self._hyperparameters = self._model.get_params()

        if self._target in self._feature_set(data_schema):
            msg = f"Target feature {self._target} is in the feature set."
            logger.error(msg)
            raise ValueError(msg)

        validation_datasets: list[tuple[str, pd.DataFrame, pd.Series]] = []
        train_df = self._memoized_load_dataset(
            data_schema,
            HashableVaexDataFrame(train_dataframe),
            (self._target,),
            name="training",
        )
        X_train = train_df[self._feature_set(data_schema)]
        y_train = train_df[self._target]
        validation_datasets.append(("train", X_train, y_train))

        if validation_dataframe is not None:
            validation_df = self._memoized_load_dataset(
                data_schema,
                HashableVaexDataFrame(validation_dataframe),
                (self._target,),
                name="validation",
            )
            X_validation = validation_df[self._feature_set(data_schema)]
            y_validation = validation_df[self._target]
            validation_datasets.append(("validation", X_validation, y_validation))

        if hyperparameters is None:
            hyperparameters = {}

        hyperparameters = {**self._hyperparameters, **hyperparameters}
        self._model.set_params(**hyperparameters)
        logger.info(
            f"Training the {self._model.__class__.__qualname__!r} model with the "
            f"following hyperparameters: \n{pp.pformat(hyperparameters)}\t"
        )

        metrics = {}
        callbacks = [lgb.record_evaluation(metrics)]
        if self._log_eval_period is not None:
            callbacks.append(lgb.log_evaluation(period=self._log_eval_period))

        self._model.fit(
            X=X_train,
            y=y_train,
            eval_set=[(X_df, y) for _, X_df, y in validation_datasets],
            eval_names=[name for name, _, _ in validation_datasets],
            eval_metric=self._eval_metric,
            feature_name=self._feature_set(data_schema),
            categorical_feature=data_schema.get_features(["categorical", "boolean"]),
            callbacks=callbacks,
        )

        return self._model, metrics

    def _transform(self, data_schema: DataSchema, dataframe: vaex.DataFrame) -> vaex.DataFrame:
        """Transforms the given dataframe using the LightGBM model.

        Will return the predictions of the model applied to the given dataframe.

        Args:
            data_schema: The data schema of the dataframe.
            dataframe: The dataframe to transform.

        Returns:
            The transformed dataframe.
        """
        dataset = self._memoized_load_dataset(data_schema, HashableVaexDataFrame(dataframe))
        df = dataframe.copy()

        logger.info("Transforming the dataset.")
        if isinstance(self._model, lgb.LGBMClassifier):
            probs: np.ndarray = self._model.predict_proba(dataset)  # type: ignore
            for i, prob in enumerate(probs.T):
                df[f"probability_{i}"] = prob
            preds: np.ndarray = probs.argmax(axis=1)
            df["prediction"] = preds
        else:
            predictions = self._model.predict(dataset)
            df["prediction"] = predictions
        return df

    def _fingerprint(self) -> Hashable:
        """Returns the fingerprint of the model.

        Appends the target feature and number of iterations to the fingerprint.

        Returns:
            The fingerprint of the model.
        """
        return (super()._fingerprint(), self._target, self._model.__class__.__qualname__)

    def _default_features(self, data_schema: DataSchema) -> tuple[str, ...]:
        """The default set of features to use for training.

        Args:
            data_schema: The data schema of the dataframes.

        Returns:
            The default set of features.
        """
        features = data_schema.get_features(["numerical", "boolean", "categorical"])
        return tuple(str(feature) for feature in features)



"""The subpackage contains the model building functionality of the `mleko` library.

The following modules are provided:
    - `base_model`: The module provides the base class for all models.
    - `lgbm_model`: The module provides functionality for building LightGBM models.
"""

from .base_model import BaseModel
from .lgbm_model import LGBMModel


__all__ = ["LGBMModel", "BaseModel"]



"""The module containing the CallableSourceFingerprinter class."""

from __future__ import annotations

import hashlib
import inspect
from typing import Callable

from .base_fingerprinter import BaseFingerprinter


class CallableSourceFingerprinter(BaseFingerprinter):
    """A fingerprinter for Callables."""

    def fingerprint(self, data: Callable) -> str:
        """Generate a fingerprint for a Python Callable.

        Note:
            The fingerprint is generated by hashing the source code of the Callable.
            A side effect of this is that the fingerprint will change if the source code
            of the Callable changes. However, any changes to variables outside of the
            Callable's scope will not affect the fingerprint.

        Args:
            data: The Callable to be fingerprinted.

        Returns:
            The fingerprint as a hexadecimal string.
        """
        fingerprint = hashlib.md5(inspect.getsource(data).encode()).hexdigest()
        return fingerprint



"""The module contains the `CSVToVaexConverter`, which converts CSV to a random-access `vaex` compatible format."""

from __future__ import annotations

import multiprocessing
from concurrent import futures
from itertools import repeat
from pathlib import Path

import pyarrow as pa
import vaex
from pyarrow import csv as arrow_csv
from tqdm.auto import tqdm

from mleko.cache.fingerprinters import CSVFingerprinter
from mleko.cache.handlers import CacheHandler
from mleko.cache.handlers.joblib_cache_handler import JOBLIB_CACHE_HANDLER
from mleko.cache.handlers.vaex_cache_handler import (
    VAEX_DATAFRAME_CACHE_HANDLER,
    read_vaex_dataframe,
    write_vaex_dataframe,
)
from mleko.dataset.data_schema import DataSchema
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr
from mleko.utils.file_helpers import clear_directory
from mleko.utils.vaex_helpers import get_column

from .base_converter import BaseConverter


logger = CustomLogger()
"""A module-level logger instance."""

V_CPU_COUNT = multiprocessing.cpu_count()
"""A module-level constant representing the total number of CPUs available on the current system."""

RESERVED_KEYWORDS = {
    "False",
    "class",
    "from",
    "or",
    "None",
    "continue",
    "global",
    "pass",
    "True",
    "def",
    "if",
    "raise",
    "and",
    "del",
    "import",
    "return",
    "as",
    "elif",
    "in",
    "try",
    "assert",
    "else",
    "is",
    "while",
    "async",
    "except",
    "lambda",
    "with",
    "await",
    "finally",
    "nonlocal",
    "yield",
    "break",
    "for",
    "not",
}
"""A module-level constant representing the reserved keywords in Python."""


def write_vaex_dataframe_with_cleanup(cache_file_path: Path, output: vaex.DataFrame) -> None:
    """Writes the results of the DataFrame conversion to a file and cleans up the cache directory.

    Args:
        cache_file_path: The path of the cache file to be written.
        output: The Vaex DataFrame to be saved in the cache file.
    """
    write_vaex_dataframe(cache_file_path, output)
    clear_directory(cache_file_path.parent, pattern="df_chunk_*.arrow")


class CSVToVaexConverter(BaseConverter):
    """A class that converts CSV to a random-access `vaex` compatible format."""

    @auto_repr
    def __init__(
        self,
        forced_numerical_columns: list[str] | tuple[str, ...] | tuple[()] = (),
        forced_categorical_columns: list[str] | tuple[str, ...] | tuple[()] = (),
        forced_boolean_columns: list[str] | tuple[str, ...] | tuple[()] = (),
        drop_columns: list[str] | tuple[str, ...] | tuple[()] = (),
        meta_columns: list[str] | tuple[str, ...] | tuple[()] = (),
        drop_rows_with_na_columns: list[str] | tuple[str, ...] | tuple[()] = (),
        na_values: list[str] | tuple[str, ...] | tuple[()] = (
            "-9998",
            "-9998.0",
            "-9999",
            "-9999.0",
            "-99",
            "-99.0",
            "nan",
            "none",
            "non",
            "Nan",
            "None",
            "Non",
            "",
            "N/A",
            "N/a",
            "unknown",
            "missing",
        ),
        true_values: list[str] | tuple[str, ...] | tuple[()] = ("t", "True", "true", "1"),
        false_values: list[str] | tuple[str, ...] | tuple[()] = ("f", "False", "false", "0"),
        downcast_float: bool = False,
        random_state: int | None = 42,
        num_workers: int = V_CPU_COUNT,
        cache_directory: str | Path = "data/csv-to-vaex-converter",
        cache_size: int = 1,
    ) -> None:
        """Initializes the `CSVToArrowConverter` with the necessary configurations and parameters.

        Args:
            forced_numerical_columns: A sequence of column names to force as numerical type.
            forced_categorical_columns: A sequence of column names to force as categorical type.
            forced_boolean_columns: A sequence of column names to force as boolean type.
            drop_columns: A sequence of column names to drop during conversion.
            meta_columns: A sequence of column names to be considered as metadata (e.g. ID or target columns).
            drop_rows_with_na_columns: A sequence of column names to drop rows with missing values.
            na_values: A sequence of strings to consider as NaN or missing values.
            true_values: A sequence of strings to consider as True values.
            false_values: A sequence of strings to consider as False values.
            downcast_float: If True, downcast float64 to float32 during conversion.
            random_state: A seed for the random number generator.
            num_workers: Number of workers to use for parallel processing.
            cache_directory: The directory where the converted files will be saved.
            cache_size: Maximum number of cache entries for the LRUCacheMixin.

        Warning:
            The `forced_numerical_columns`, `forced_categorical_columns`, `forced_boolean_columns`, and `drop_columns`
            parameters are mutually exclusive. Meaning, a column cannot be specified in more than one of these
            parameters. If a column is specified in more than one of these parameters, the last parameter will be used
            and the previous ones will be ignored.

        Example:
            >>> import vaex
            >>> from mleko.dataset.convert import CSVToArrowConverter
            >>> converter = CSVToArrowConverter(
            ...     forced_numerical_columns=["x"],
            ...     forced_categorical_columns=["y"],
            ...     forced_boolean_columns=["z"],
            ...     drop_columns=["a"],
            ...     na_values=["-9999"],
            ...     true_values=["t"],
            ...     false_values=["f"],
            ...     downcast_float=True,
            ...     random_state=42,
            ...     num_workers=4,
            ...     cache_size=1,
            ... )
            >>> df = converter.convert(["data.csv"])
        """
        super().__init__(cache_directory, cache_size)
        self._forced_numerical_columns = tuple(forced_numerical_columns)
        self._forced_categorical_columns = tuple(forced_categorical_columns)
        self._forced_boolean_columns = tuple(forced_boolean_columns)
        self._drop_columns = tuple(drop_columns)
        self._meta_columns = tuple(meta_columns)
        self._drop_rows_with_na_columns = tuple(drop_rows_with_na_columns)
        self._na_values = tuple(na_values)
        self._true_values = tuple(true_values)
        self._false_values = tuple(false_values)
        self._downcast_float = downcast_float
        self._num_workers = num_workers
        self._random_state = random_state

    def convert(
        self,
        file_paths: list[Path] | list[str],
        cache_group: str | None = None,
        force_recompute: bool = False,
        disable_cache: bool = False,
    ) -> tuple[DataSchema, vaex.DataFrame]:
        """Converts a list of CSV files to Arrow format and returns a `vaex` dataframe joined from the converted data.

        The method takes care of caching, and results will be reused accordingly unless `force_recompute`
        is set to True. The resulting dataframe is a `vaex` DataFrame joined from the converted data.
        The conversion is done in chunks to optimize parallel processing.

        Note:
            Will read the first `100,000/len(file_paths)` rows of each file to determine if the file is the same as the
            one in the cache. If the file is the same, the cache will be used. Otherwise, the file will be converted
            and the cache will be updated.

        Args:
            file_paths: A list of file paths to be converted.
            cache_group: The cache group to use.
            force_recompute: If set to True, forces recomputation and ignores the cache.
            disable_cache: If set to True, disables the cache.

        Returns:
            The resulting dataframe with the combined converted data.
        """
        ds, df = self._cached_execute(
            lambda_func=lambda: self._convert(file_paths),
            cache_key_inputs=[
                self._forced_numerical_columns,
                self._forced_categorical_columns,
                self._forced_boolean_columns,
                self._drop_columns,
                self._meta_columns,
                self._drop_rows_with_na_columns,
                self._na_values,
                self._true_values,
                self._false_values,
                self._downcast_float,
                (file_paths, CSVFingerprinter(n_rows=100_000 // len(file_paths))),
            ],
            cache_group=cache_group,
            force_recompute=force_recompute,
            cache_handlers=[
                JOBLIB_CACHE_HANDLER,
                CacheHandler(
                    writer=write_vaex_dataframe_with_cleanup,
                    reader=read_vaex_dataframe,
                    suffix=VAEX_DATAFRAME_CACHE_HANDLER.suffix,
                    can_handle_none=False,
                ),
            ],
            disable_cache=disable_cache,
        )
        return ds, df

    @staticmethod
    def _convert_csv_file_to_arrow(
        file_path: Path | str,
        output_directory: Path,
        forced_numerical_columns: tuple[str, ...],
        forced_categorical_columns: tuple[str, ...],
        forced_boolean_columns: tuple[str, ...],
        drop_columns: tuple[str, ...],
        na_values: tuple[str, ...],
        true_values: tuple[str, ...],
        false_values: tuple[str, ...],
        downcast_float: bool,
    ) -> None:
        """Converts a single CSV file to Arrow format using the provided options and saves it to the output directory.

        This operation is done in chunks to optimize parallel processing. The resulting dataframe is saved in the
        output directory with the given suffix.

        Args:
            file_path: The path of the CSV file to be converted.
            output_directory: The directory where the converted file should be saved.
            forced_numerical_columns: A sequence of column names to be forced to numerical type.
            forced_categorical_columns: A sequence of column names to be forced to categorical type.
            forced_boolean_columns: A sequence of column names to be forced to boolean type.
            drop_columns: A sequence of column names to be dropped from the dataframe.
            na_values: A sequence of values to be considered as NaN.
            true_values: A sequence of values to be considered as True.
            false_values: A sequence of values to be considered as False.
            downcast_float: If set to True, downcasts float64 to float32.
        """
        file_path = Path(file_path)

        float_type = "float64"
        if downcast_float:
            float_type = "float32"

        dtypes = {}
        for col in forced_numerical_columns:
            dtypes[col] = float_type
        for col in forced_categorical_columns:
            dtypes[col] = "string"
        for col in forced_boolean_columns:
            dtypes[col] = "boolean"

        df_chunk = vaex.from_csv_arrow(
            file_path,
            read_options=arrow_csv.ReadOptions(use_threads=True),
            convert_options=arrow_csv.ConvertOptions(
                column_types=dtypes,
                null_values=na_values,
                true_values=true_values,
                false_values=false_values,
                strings_can_be_null=True,
                quoted_strings_can_be_null=True,
                timestamp_parsers=[
                    arrow_csv.ISO8601,
                    "%Y-%m-%d %H:%M:%S",
                    "%Y-%m-%d %H:%M:%S.%f",
                    "%Y-%m-%dT%H:%M:%S",
                    "%Y-%m-%dT%H:%M:%S.%f",
                ],
            ),
        ).drop(drop_columns)

        for column_name in df_chunk.get_column_names():
            if get_column(df_chunk, column_name).dtype in (pa.date32(), pa.date64()):
                df_chunk[column_name] = get_column(df_chunk, column_name).astype("datetime64[s]")

        output_path = output_directory / f"df_chunk_{file_path.stem}.arrow"
        df_chunk.export(output_path, chunk_size=100_000, parallel=False)
        df_chunk.close()

    def _convert(self, file_paths: list[Path] | list[str]) -> tuple[DataSchema, vaex.DataFrame]:
        """Converts a list of CSV files to Arrow format using parallel processing.

        Chunks of files are processed in parallel and saved in the output directory.

        Args:
            file_paths: A list of file paths to be converted.

        Returns:
            A DataFrame containing the merged chunks.
        """
        with tqdm(total=len(file_paths), desc="Converting CSV files") as pbar:
            with futures.ProcessPoolExecutor(max_workers=min(self._num_workers, len(file_paths))) as executor:
                for _ in executor.map(
                    CSVToVaexConverter._convert_csv_file_to_arrow,
                    file_paths,
                    repeat(self._cache_directory),
                    repeat(self._forced_numerical_columns),
                    repeat(self._forced_categorical_columns),
                    repeat(self._forced_boolean_columns),
                    repeat(self._drop_columns),
                    repeat(self._na_values),
                    repeat(self._true_values),
                    repeat(self._false_values),
                    repeat(self._downcast_float),
                ):
                    pbar.update(1)

        logger.info("Finished converting CSV files to Vaex format.")
        df: vaex.DataFrame = vaex.open(self._cache_directory / "df_chunk_*.arrow")
        logger.info("Renaming columns with non-compatible names or reserved keywords.")
        for column_name in df.get_column_names():
            if column_name in RESERVED_KEYWORDS:
                logger.warning(f"Renaming column {column_name!r} to '_{column_name}'")
                df.rename(column_name, f"_{column_name}")

            if column_name == "":
                logger.warning(f"Renaming column {column_name!r} to '_empty'")
                df.rename(column_name, "_empty")

        for column_name in df.get_column_names(dtype=pa.null()):
            df[column_name] = get_column(df, column_name).astype("string")

        if self._drop_rows_with_na_columns:
            df = df.dropna(column_names=self._drop_rows_with_na_columns)

        ds = DataSchema(
            numerical=df.get_column_names(dtype="numeric"),
            categorical=df.get_column_names(dtype="string"),
            boolean=df.get_column_names(dtype="bool"),
            datetime=df.get_column_names(dtype="datetime"),
            timedelta=df.get_column_names(dtype="timedelta"),
        )
        ds.drop_features(self._meta_columns)

        for column_name in df.get_column_names(dtype="bool"):
            df[column_name] = get_column(df, column_name).astype("int8")

        logger.info("Merging chunks into a single DataFrame.")
        return ds, df



"""This module contains the CacheHandler for serializing and deserializing data using joblib."""

from pathlib import Path
from typing import Any

import joblib

from .base_cache_handler import CacheHandler


def write_joblib(cache_file_path: Path, output: Any) -> None:
    """Writes the given data to a cache file at the specified path, serializing it using joblib.

    Args:
        cache_file_path: A Path object representing the location where the cache file should be saved.
        output: The data to be serialized and saved to the cache file.
    """
    joblib.dump(output, cache_file_path)


def read_joblib(cache_file_path: Path) -> Any:
    """Reads the cache file from the specified path and returns the deserialized data.

    Args:
        cache_file_path: A Path object representing the location of the cache file.

    Returns:
        The deserialized data stored in the cache file.
    """
    return joblib.load(cache_file_path)


JOBLIB_CACHE_HANDLER = CacheHandler(write_joblib, read_joblib, "joblib", can_handle_none=True)
"""A CacheHandler for Python objects using joblib."""



"""Module for data conversion as a part of the pipeline process.

This module contains the `ConvertStep` class which is a specialized pipeline step for handling data format
conversion. It uses the provided `BaseDataConverter` for converting the data into the desired format.
"""

from __future__ import annotations

from pathlib import Path
from typing import List, Union

from typing_extensions import TypedDict, TypeVar

from mleko.dataset.convert import BaseConverter
from mleko.pipeline.data_container import DataContainer
from mleko.pipeline.pipeline_step import PipelineStep
from mleko.utils.custom_logger import CustomLogger
from mleko.utils.decorators import auto_repr


logger = CustomLogger()
"""The logger for the module."""

TypedDictType = TypeVar("T", bound=TypedDict)  # type: ignore
"""Type variable for TypedDict type annotations."""


class ConvertStepInputType(TypedDict):
    """The input type of the ConvertStep."""

    file_paths: Union[str, List[Path], List[str]]
    """List of file paths or the key identifying the list of file paths to be converted."""


class ConvertStepOutputType(TypedDict):
    """The output type of the ConvertStep."""

    data_schema: str
    """The key for the DataSchema after conversion."""

    dataframe: str
    """The key for the DataFrame after conversion."""


class ConvertStep(PipelineStep):
    """Pipeline step that manages data conversion from one format to another."""

    _inputs: ConvertStepInputType
    _outputs: ConvertStepOutputType

    @auto_repr
    def __init__(
        self,
        converter: BaseConverter,
        inputs: ConvertStepInputType,
        outputs: ConvertStepOutputType,
        cache_group: str | None = None,
    ) -> None:
        """Initialize the ConvertStep with the specified data converter.

        Args:
            converter: The DataConverter responsible for handling data format conversion.
            inputs: A dictionary of input keys following the `ConvertStepInputType` schema.
            outputs: A dictionary of output keys following the `ConvertStepOutputType` schema.
            cache_group: The cache group to use.
        """
        super().__init__(inputs, outputs, cache_group)
        self._converter = converter

    def execute(self, data_container: DataContainer, force_recompute: bool) -> DataContainer:
        """Perform data format conversion using the configured converter.

        Args:
            data_container: Contains a list of file Paths to be converted.
            force_recompute: Whether to force the step to recompute its output, even if it already exists.

        Raises:
            ValueError: If the input data types are invalid.

        Returns:
            A DataContainer containing the result.
        """
        file_paths = self._validate_and_get_input(self._inputs["file_paths"], list, data_container)
        if not all(isinstance(e, Path) or isinstance(e, str) for e in file_paths):
            msg = "Invalid data type: {type(file_paths)}. Expected List[Path] or List[str]."
            logger.error(msg)
            raise ValueError(msg)

        data_schema, dataframe = self._converter.convert(file_paths, self._cache_group, force_recompute)
        data_container.data[self._outputs["data_schema"]] = data_schema
        data_container.data[self._outputs["dataframe"]] = dataframe
        return data_container

    def _get_input_model(self) -> type[ConvertStepInputType]:
        """Get the input type for the ConvertStep.

        Returns:
            Input type for the ConvertStep.
        """
        return ConvertStepInputType

    def _get_output_model(self) -> type[ConvertStepOutputType]:
        """Get the output type for the ConvertStep.

        Returns:
            Output type for the ConvertStep.
        """
        return ConvertStepOutputType
